package de.uni.leipzig.informatik.pacosy.cophylogeny.solver.cplex;

import ilog.concert.IloException;
import ilog.concert.IloIntExpr;
import ilog.concert.IloIntVar;
import ilog.concert.IloNumExpr;
import ilog.concert.IloNumVar;
import ilog.cplex.IloCplex;
import ilog.cplex.IloCplex.UnknownObjectException;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PipedOutputStream;
import java.util.Collection;

import de.uni.leipzig.informatik.pacosy.cophylogeny.Parameters;
import de.uni.leipzig.informatik.pacosy.datastructure.dao.Edge;
import de.uni.leipzig.informatik.pacosy.datastructure.dao.Node;
import de.uni.leipzig.informatik.pacosy.datastructure.dao.Tanglegramm;
import de.uni.leipzig.informatik.pacosy.datastructure.dao.Tree;
import de.uni.leipzig.informatik.pacosy.datastructure.util.NodeUtil;


public class CplexCophylogenySolver extends AbstractCplexSolver {
	
	// input data
	private Tanglegramm tanglegramm;
	// CPLEX solver
	private IloCplex solver;
	// objective function
	private IloNumExpr objFn;
	// factor variables
	private IloNumVar[] fp;
	private IloNumVar[] fh;
	// time variables
	private IloIntVar[] t2p;
	private IloIntVar[] t2h;
	// association variables
	private IloIntVar[][] a;
	// subtree association variables
	private IloIntVar[][] s;
	// exclusive association variables
	private IloIntVar[][] e;
	// cospeciation variables
	private IloIntVar[][] c;
	// horizontal transfer variables
	private IloIntVar[][] h;
	
	// no of hosts
	private Tree hostTree;
	private int noHosts;
	// no of parasites
	private Tree parasiteTree;
	private int noParasites;
	
	private Parameters parameters;
	
	double blFactor=1;
	long time;
		
	public CplexCophylogenySolver(Tanglegramm tanglegramm, Parameters parameters){
		this.tanglegramm = tanglegramm;
		this.hostTree = tanglegramm.getTrees().get("HOST");
		this.noHosts = this.hostTree.getSize();
		this.parasiteTree = tanglegramm.getTrees().get("PARASITE");
		this.noParasites = parasiteTree.getSize();
		this.parameters = parameters;
	}
	
	public boolean solve() throws IloException, IOException{
		
		double lb = this.parameters.getLowerBound();
		double ub = this.parameters.getUpperBound();
		boolean minHostSwitches = this.parameters.isHostswitch();
		boolean minAssociations = this.parameters.isAssociations();

		this.time = System.currentTimeMillis();
		int maxOutdegree = Math.min(NodeUtil.getMaxOutdegree(parasiteTree.getRoot()), NodeUtil.getMaxOutdegree(hostTree.getRoot()));
		double maxBranchLength = Math.min(NodeUtil.getMaxBranchLength(parasiteTree), NodeUtil.getMaxBranchLength(hostTree));
		// we scale all branch lengths to a size such that we can use integer numbers for the timestamps
		// the longest branch should have a length of 4*(noHosts+noParasites), i.e., for each node there are on average
		// 4 possible exclusive integer time points if all the coevolution is mapped onto this longest branch 
		this.blFactor = Math.pow(10, Math.ceil(Math.log10((4*(noHosts+noParasites))/(lb*maxBranchLength))));
		
		// initialize the CLPEX instance
		this.solver = new IloCplex();
		this.solver.setName("CplexCophylogenySolver");
		
		// initialize factor variable
		this.fh = solver.numVarArray(this.noHosts, lb, ub);
		this.fp = solver.numVarArray(this.noParasites, lb, ub);
		
		double maxTime = ub*Math.max(NodeUtil.getMaxDepth(hostTree.getRoot(), hostTree), NodeUtil.getMaxDepth(parasiteTree.getRoot(), parasiteTree));
		
		// initialize time variable
		this.t2h = solver.intVarArray(this.noHosts, 0, (int)Math.ceil(maxTime*blFactor));
		this.t2p = solver.intVarArray(this.noParasites, 0,  (int)Math.ceil(maxTime*blFactor));
		
		// initialize association variable
		this.a = new IloIntVar[this.noParasites][this.noHosts];
		for (int i=0; i<this.noParasites; i++){
			this.a[i] = solver.boolVarArray(this.noHosts);
			if (this.parameters.isSingle()){
				solver.addEq(solver.sum(this.a[i]), 1);
			}
			else {
				solver.addGe(solver.sum(this.a[i]), 1);
			}
		}
		
		// initialize subtree association variable
		this.s = new IloIntVar[this.noParasites][this.noHosts];
		for (int i=0; i<this.noParasites; i++){
			this.s[i] = solver.boolVarArray(this.noHosts);
		}
		
		// initialize exclusive association variable
		this.e = new IloIntVar[this.noParasites][this.noHosts];
		for (int i=0; i<this.noParasites; i++){
			this.e[i] = solver.boolVarArray(this.noHosts);
		}
		
		// initialize cospeciation variable
		this.c = new IloIntVar[this.noParasites][this.noHosts];
		for (int i=0; i<this.noParasites; i++){
			this.c[i] = solver.intVarArray(this.noHosts, 0, maxOutdegree);
		}
		
		if (minHostSwitches){
			// initialize hostswitch variable
			this.h = new IloIntVar[this.noParasites][this.noHosts];
			for (int i=0; i<this.noParasites; i++){
				this.h[i] = solver.boolVarArray(this.noHosts);
			}
		}
		
		Collection<Node> hostNodes = this.tanglegramm.getTrees().get("HOST").getNodes();
		Collection<Node> parasiteNodes = this.tanglegramm.getTrees().get("PARASITE").getNodes();
		
		// set input associations and no cospeciation on leaves
		for (Node parasite : parasiteNodes){
			for (Node host : hostNodes){
				int i = parasite.getNumber();
				int j = host.getNumber();
				if (parasite.getAssociations().containsKey(host) && parasite.getAssociations().get(host)>=0.5){
					solver.addEq(a[i][j], 1);
				}
				else if (parasite.isLeaf() && host.isLeaf()){
					solver.addEq(a[i][j], 0);
				}
				
				if (parasite.isLeaf() && !host.isLeaf()){
					solver.addEq(a[i][j], 0);
				}
				
				if (parasite.isLeaf() || host.isLeaf()){
					solver.addEq(c[i][j], 0);
				}
			}
		}
		if (minHostSwitches){
			int i = this.parasiteTree.getRoot().getNumber();
			for (Node host : hostNodes){
				int j = host.getNumber();
				solver.addEq(h[i][j], 0);
			}
			int j = this.hostTree.getRoot().getNumber();
			for (Node parasite : parasiteNodes){
				i = parasite.getNumber();
				solver.addEq(h[i][j], 0);
			}
		}
		
		// time constraints for child nodes t_i'-0.5 <= t_i+f_i*bl(i) <= t_i' + 0.5
		for (Node parasite : parasiteNodes){
			int i = parasite.getNumber();
			if (parasite.isLeaf()){
				solver.addEq(this.t2p[i], 0);
			}
			else {
				for (Node child : parasite.getChildren()){
					Edge edge = this.parasiteTree.getEdge(parasite.getId(), child.getId());
					int k = child.getNumber();
					solver.addLe(solver.sum(t2p[i], -0.5), solver.sum(t2p[k], solver.prod(blFactor*edge.getLength(), fp[k])));
					solver.addGe(solver.sum(t2p[i],  0.5), solver.sum(t2p[k], solver.prod(blFactor*edge.getLength(), fp[k])));
				}
			}
		}
		for (Node host : hostNodes){
			int j = host.getNumber();
			if (host.isLeaf()){
				solver.addEq(this.t2h[j], 0);
			}
			else {
				for (Node child : host.getChildren()){
					Edge edge = this.hostTree.getEdge(host.getId(), child.getId());
					int l = child.getNumber();
					solver.addLe(solver.sum(t2h[j], -0.5), solver.sum(t2h[l], solver.prod(blFactor*edge.getLength(), fh[l])));
					solver.addGe(solver.sum(t2h[j],  0.5), solver.sum(t2h[l], solver.prod(blFactor*edge.getLength(), fh[l])));
				}
			}
		}
		
		// t2h, t2p time constraints in case of association a_{ij}=1 -> t_{i}-t_j >= 0 and t_{j'}-t_i >= 1
		for (Node parasite : parasiteNodes){
			for (Node host : hostNodes){
				int i = parasite.getNumber();
				int j = host.getNumber();
				solver.addGe(solver.sum(solver.prod(maxTime*blFactor, solver.sum(1, solver.negative(a[i][j]))),t2p[i],solver.negative(t2h[j])), 0);
				if (host.getParent()!=null){
					int jp = host.getParent().getNumber();
					solver.addGe(solver.sum(solver.prod(maxTime*blFactor, solver.sum(1, solver.negative(a[i][j]))),t2h[jp],solver.negative(t2p[i])), 1);
				}
			}
		}
		
		// fix s_{ij} = 1 iff \sum_{j_l \in desc(j)} a_{ij_l} >= 1
		// fix s_{ij} = 1 iff a_{ij} + \sum_{j_l \in child(j)} s_{ij_l} >= 1
		for (Node parasite : parasiteNodes){
			for (Node host : hostNodes){
				int i = parasite.getNumber();
				int j = host.getNumber();
				IloIntExpr expr = a[i][j];
				for (Node child : host.getChildren()){
					int l = child.getNumber();
					expr = solver.sum(expr, s[i][l]);
				}
				solver.addLe(s[i][j], expr);
				solver.addGe(solver.prod(s[i][j],host.getChildren().size()+1), expr);
			}
		}
		
		// ensure backtraceability: if a_{ij} = 1 then \sum_{i_k \in child(i) s_{i_kj} >= 1
		for (Node parasite : parasiteNodes){
			if (!parasite.isLeaf()){
				for (Node host : hostNodes){
					int i = parasite.getNumber();
					int j = host.getNumber();
					IloIntExpr expr = solver.constant(0);
					for (Node child : parasite.getChildren()){
						int k = child.getNumber();
						expr = solver.sum(expr, s[k][j]);
					}
					solver.addLe(a[i][j], expr);
				}
			}
		}
		
		// if a_{ij}=1 then force no subtree association of i within a child of j, i.e., a_{ij}=1 -> \sum_{j_l \in child(j)} s_{ij_l} = 0
		// I don't know why we needed this previously, but the constraint is (hopefully) assured by the time constraints in case of an association
//		for (Node parasite : parasiteNodes){
//			for (Node host : hostNodes){
//				int n = host.getChildren().size();
//				int i = parasite.getNumber();
//				int j = host.getNumber();
//				IloIntExpr expr = solver.constant(0);
//				for (Node child : host.getChildren()){
//					int l = child.getNumber();
//					expr = solver.sum(expr, s[i][l]);
//				}
//				solver.addGe(solver.prod(n, solver.sum(1,solver.negative(a[i][j]))), expr);
//			}
//		}
		
		// fix e_{ij} exclusive sibling association: e_{ij} = 1 iff s_{ij}=1 and \sum_{i_k \neq i} \sum_{j_l \neq j} s_{i_kj_l}=0
		for (Node parasite : parasiteNodes){
			if (parasite.getParent()!=null){
				for (Node host : hostNodes){
					if (host.getParent()!=null){
						int n = parasite.getParent().getChildren().size()+host.getParent().getChildren().size()+1;
						int i = parasite.getNumber();
						int j = host.getNumber();
						IloIntExpr expr = solver.sum(solver.prod(n, e[i][j]),solver.sum(1, solver.negative(s[i][j])));
						for (Node parasiteSibling : parasite.getParent().getChildren()){
							int k = parasiteSibling.getNumber();
							if (i!=k){
								expr = solver.sum(expr, s[k][j]);
							}
						}
						for (Node hostSibling : host.getParent().getChildren()){
							int l = hostSibling.getNumber();
							if (j!=l){
								expr = solver.sum(expr, s[i][l]);
							}
						}
						solver.addRange(1, expr, n);
					}
				}
			}
		}
		
		// fix c_{ij} number of cospeciations: if i and j are associated then c_{ij} = max(0, exclusive child associations-1); otherwise c_{ij} = 0
		for (Node parasite : parasiteNodes){
			for (Node host : hostNodes){
				if (!parasite.isLeaf() && !host.isLeaf()){
					int i = parasite.getNumber();
					int j = host.getNumber();
					IloIntExpr exprExclusive = solver.constant(0);
					for (Node parasiteChild : parasite.getChildren()){
						for (Node hostChild : host.getChildren()){
							int k = parasiteChild.getNumber();
							int l = hostChild.getNumber();
							exprExclusive = solver.sum(exprExclusive, e[k][l]);
						}
					}
					IloIntExpr exprChild = solver.constant(0);
					for (Node parasiteChild : parasite.getChildren()){
						int k = parasiteChild.getNumber();
						exprChild = solver.sum(exprChild, a[k][j]);
					}
					
					IloIntExpr exprNot = solver.sum(1, solver.negative(a[i][j]));
					// if a_{ij} = 1 and \sum_{i_k \in child(i)} a_{i_kj} = 0 then c_{ij} + 1 >= \sum_{i_k} \sum_{j_l} e_{i_kj_l}
					solver.addGe(solver.sum(solver.prod(maxOutdegree, exprNot),solver.prod(maxOutdegree, exprChild),c[i][j],solver.negative(exprExclusive)),-1);
					// if a_{ij} = 1 and \sum_{i_k \in child(i)} a_{i_kj} = 0 then c_{ij} <= \sum_{i_k} \sum_{j_l} e_{i_kj_l}
					solver.addGe(solver.sum(solver.prod(maxOutdegree, exprNot),solver.prod(maxOutdegree, exprChild),exprExclusive,solver.negative(c[i][j])),0);
					// if a_{ij} = 1 and \sum_{i_k \in child(i)} a_{i_kj} = 0 and \sum\sum e_{i_kj_l} >= 1 then c_{ij} + 1 = \sum_{i_k} \sum_{j_l} e_{i_kj_l}
					solver.addLe(0, solver.sum(solver.prod(maxOutdegree, exprNot),solver.prod(maxOutdegree, exprChild),solver.sum(0.5d,solver.negative(exprExclusive)),solver.prod(maxOutdegree, solver.sum(exprExclusive, solver.negative(c[i][j])))));
					
					// if a_{ij} = 0 then c_{ij} = 0
					solver.addGe(solver.prod(maxOutdegree,a[i][j]),c[i][j]);
					for (Node parasiteChild : parasite.getChildren()){
						int k = parasiteChild.getNumber();
						// for all i_k \in child(i) if a_{i_kj} = 1 then c_{ij} = 0
						solver.addGe(solver.prod(maxOutdegree,solver.sum(1,solver.negative(a[k][j]))),c[i][j]);
					}
				}
			}
		}
		
		if (minHostSwitches){
			// fix h_{ij} number of hostswitches: if i and j are associated then h_{ij} = \sum_{i_k \in child(i)} s_{i_kj}; otherwise c_{ij} = 0
			for (Node parasite : parasiteNodes){
				for (Node host : hostNodes){
					if (parasite.getParent()!=null && host.getParent()!=null){
						int i = parasite.getNumber();
						int j = host.getNumber();
						IloIntExpr exprSubtree = solver.constant(0);
						Node parasiteParent = parasite.getParent();
						int ip = parasiteParent.getNumber();
						Node hostParent = host;
						while (hostParent != null){
							int jp = hostParent.getNumber();
							exprSubtree = solver.sum(exprSubtree, a[ip][jp]);
							hostParent = hostParent.getParent();
						}
						
						// if a_{ij} = 1 and \sum_{j_l \in ancestor(j)} s_{i'j_l} = 0 then h_{ij} = 1
						solver.addGe(solver.sum(solver.sum(1, solver.negative(a[i][j])), exprSubtree, h[i][j]), 1);
						
						// if a_{ij} = 0 then h_{ij} = 0
						solver.addGe(a[i][j],h[i][j]);
						hostParent = host;
						while (hostParent != null){
							int jp = hostParent.getNumber();
							// if s_{i'j_l} = 1 then h_{ij} = 0
							solver.addGe(solver.sum(1,solver.negative(a[ip][jp])),h[i][j]);
							hostParent = hostParent.getParent();
						}
					}
				}
			}
		}
		
		// time constraint for cospeciation: c_{ij} >= 1 -> t_i = t_j
		for (Node parasite : parasiteNodes){
			for (Node host : hostNodes){
				int i = parasite.getNumber();
				int j = host.getNumber();
				solver.add(solver.ifThen(solver.ge(c[i][j], 1), solver.le(t2p[i], t2h[j])));
			}
		}
		
		this.objFn = solver.constant(0);
		for (Node parasite : parasiteNodes){
			int i = parasite.getNumber();
			this.objFn = solver.sum(this.objFn, solver.negative(solver.sum(c[i])));
		}
		
		if (minHostSwitches){
			int factor = this.noHosts * this.noParasites;
			this.objFn = solver.prod(this.objFn, factor);
			for (Node parasite : parasiteNodes){
				int i = parasite.getNumber();
				this.objFn = solver.sum(this.objFn, solver.sum(h[i]));
			}
		}
		
		if (minAssociations){
			int factor = this.noHosts * this.noParasites;
			this.objFn = solver.prod(this.objFn, factor);
			for (Node parasite : parasiteNodes){
				int i = parasite.getNumber();
				this.objFn = solver.sum(this.objFn, solver.sum(a[i]));
			}
		}
		
		solver.addMinimize(this.objFn);
		if (!this.parameters.isVerbose()){
			solver.setOut(new PipedOutputStream());
			solver.setWarning(new PipedOutputStream());
		}
		
		// set priority and branching direction of m variables
//		for (Node parasite : parasiteNodes){
//			for (Node host : hostNodes){
//				int i = parasite.getNumber();
//				int j = host.getNumber();
//				solver.setPriority(a[i][j], 1);
//			}
//		}

		// solve
		solver.setParam(IloCplex.DoubleParam.SolnPoolAGap, 0.0);
		solver.setParam(IloCplex.DoubleParam.SolnPoolGap, 0.0);
		solver.setParam(IloCplex.DoubleParam.EpAGap, 0.0);
		solver.setParam(IloCplex.DoubleParam.EpGap, 0.0);
		boolean ret = solver.solve();
		
		if (ret) {
			int solutions = 1;
			writeFile("", this.parameters.getOutput(), false);
			printSolution(solutions);
			this.time = System.currentTimeMillis();
			int bestObjectiveValue = (int) (solver.getObjValue()+0.5);
			while (this.parameters.getNoSolutions() < 0 || solutions < this.parameters.getNoSolutions()){
				IloIntExpr exprSolution = solver.constant(0);
				for (IloIntVar[] ai : a){
					double[] aValues = solver.getValues(ai);
					for (int j=0; j<ai.length; j++){
						if (aValues[j]<0.5){
							exprSolution = solver.sum(exprSolution, ai[j]);
						}
						else{
							exprSolution = solver.sum(exprSolution, solver.sum(1, solver.negative(ai[j])));
						}
					}
				}
				
				IloIntExpr exprCospeciations = solver.constant(0);
				IloIntExpr exprHostswitches = solver.constant(0);
				IloIntExpr exprAssociations = solver.constant(0);
				int cospeciations = 0;
				int hostswitches = 0;
				int associations = 0;
				for (int i=0; i<this.noParasites; i++){
					double cValues[] = solver.getValues(c[i]);
					double hValues[] = new double[this.noHosts];
					double aValues[] = new double[this.noHosts];
					if (minHostSwitches){
						hValues = solver.getValues(h[i]);
					}
					if (minAssociations){
						aValues = solver.getValues(a[i]);
					}
					for (int j=0; j<this.noHosts; j++){
						cospeciations += Math.round(cValues[j]);
						if (minHostSwitches){
							hostswitches += hValues[j]<=0.5?0:1;
						}
						if (minAssociations){
							associations += aValues[j]<=0.5?0:1;
						}
					}
					exprCospeciations = solver.sum(exprCospeciations, solver.sum(c[i]));
					if (minHostSwitches){
						exprHostswitches = solver.sum(exprHostswitches, solver.sum(h[i]));
					}
					if (minAssociations){
						exprAssociations = solver.sum(exprAssociations, solver.sum(a[i]));
					}
				}
				
				solver.addGe(exprSolution, 1);
				solver.addEq(exprCospeciations, cospeciations);
				if (minHostSwitches){
					solver.addEq(exprHostswitches, hostswitches);
				}
				if (minAssociations){
					solver.addEq(exprAssociations, associations);
				}

				if (!solver.solve() || bestObjectiveValue>(int) (solver.getObjValue()+0.5)){
					break;
				}
				else{
					solutions++;
					System.out.println();
					printSolution(solutions);
					this.time = System.currentTimeMillis();
					
				}
			}
			if(solutions>1){
				System.out.print("#" + solutions + " solution(s) found\n");
			}
		}
		else {
			System.out.print("#No solution found.\n");
		}
		
		// free memory
		solver.end();
		
		return ret;

	}
	

	/**
	 * @return the current solution, i.e. the values of cVariables as string
	 * @throws UnknownObjectException
	 * @throws IloException
	 * @throws IOException 
	 */
	private void printSolution(int solutions) throws UnknownObjectException, IloException, IOException {
		String output = "";
		boolean minHostSwitches = this.parameters.isHostswitch();
		String parasites[] = new String[this.noParasites];
		String hosts[] = new String[this.noHosts];
		for (Node parasiteNode : this.tanglegramm.getTrees().get("PARASITE").getNodes()){
			parasites[parasiteNode.getNumber()]=parasiteNode.getId();
		}
		for (Node hostNode : this.tanglegramm.getTrees().get("HOST").getNodes()){
			hosts[hostNode.getNumber()]=hostNode.getId();
		}
		output += "#Solution: "+solutions+"\n";
		time = System.currentTimeMillis() - time;
		System.out.print("Duration in milliseconds:\n" + time + "\n");
		output += "#Duration in milliseconds:\n" + time + "\n";
		
		output += "#Adjusted parasite tree:\n";
		justifyBranchlengths(this.parasiteTree.getRoot(), this.parasiteTree, this.fp, this.t2p);
		output += this.parasiteTree.getScaledTree(this.parasiteTree.getRoot())+"\n";
		output += "#Adjusted host tree:\n";
		justifyBranchlengths(this.hostTree.getRoot(), this.hostTree, this.fh, this.t2h);
		output += this.hostTree.getScaledTree(this.hostTree.getRoot())+"\n";
		
		System.out.print("#EVENTS" + "\n");
		output += "#EVENTS" + "\n";
		int cospeciations = 0;
		int hostswitches = 0;
		for (int i=0; i<this.noParasites; i++){
			double cValues[] = solver.getValues(c[i]);
			double hValues[] = new double[this.noHosts];
			if (minHostSwitches){
				hValues = solver.getValues(h[i]);
			}
			for (int j=0; j<this.noHosts; j++){
				cospeciations += Math.round(cValues[j]);
				if (minHostSwitches){
					hostswitches += hValues[j]<=0.5?0:1;
				}
			}
		}
		System.out.print(cospeciations + " ? ? ");
		output += cospeciations + " ? ? ";
		if (minHostSwitches){
			System.out.println(hostswitches);
			output += hostswitches+"\n";
		}
		else{
			System.out.print("?\n");
			output += "?\n";
		}
		
		System.out.print("#RECONSTRUCTION");
		output += "#RECONSTRUCTION";
		for (int i=0; i<this.noParasites; i++){
			double aValues[] = solver.getValues(a[i]);
			for (int j=0; j<this.noHosts; j++){
				if (aValues[j]>0.5){
					System.out.print("\n"+parasites[i]+"\t"+hosts[j]);
					output += "\n"+parasites[i]+"\t"+hosts[j];
				}
			}
		}
		System.out.println();
		output += "\n";
		
		output += "#COSPECIATIONS";
		for (int i=0; i<this.noParasites; i++){
			double cValues[] = solver.getValues(c[i]);
			for (int j=0; j<this.noHosts; j++){
				if (cValues[j]>0.5){
					output += "\n"+parasites[i]+"\t"+hosts[j]+"\t"+((int)Math.round(cValues[j]));
				}
			}
		}
		output += "\n";
		
		if (minHostSwitches){
			output += "#HOSTSWITCHES";
			for (int i=0; i<this.noParasites; i++){
				double hValues[] = solver.getValues(h[i]);
				for (int j=0; j<this.noHosts; j++){
					if (hValues[j]>0.5){
						output += "\n"+parasites[i]+"\t"+hosts[j];
					}
				}
			}
			output += "\n";
		}
		
		output += "\n";
		writeFile(output, this.parameters.getOutput(), true);
		
		if (this.parameters.isVerbose()){
			
			System.out.print("#Times/Factors:\n");
			double tpValues[] = solver.getValues(t2p);
			for (int i=0; i<this.noParasites; i++){
				System.out.print(parasites[i]+"\t");
			}
			System.out.println();
			for (int i=0; i<this.noParasites; i++){
				System.out.print(tpValues[i]+"\t");
			}
			System.out.println();
			for (int i=0; i<this.noParasites; i++){
				if (i==0){
					System.out.print("-\t");
				}
				else{
					System.out.print(solver.getValue(fp[i])+"\t");
				}
			}
			System.out.println();
			
			double thValues[] = solver.getValues(t2h);
			for (int j=0; j<this.noHosts; j++){
				System.out.print(hosts[j]+"\t");
			}
			System.out.println();
			for (int i=0; i<this.noHosts; i++){
				System.out.print(thValues[i]/this.blFactor+"\t");
			}
			System.out.println();
			for (int i=0; i<this.noHosts; i++){
				if (i==0){
					System.out.print("-\t");
				}
				else{
					System.out.print(solver.getValue(fh[i])+"\t");
				}
			}
			System.out.println();
			
			
			System.out.print("#ASSOCIATION MATRIX:\n");
			int associations = 0;
			System.out.print("\t");
			for (int j=0; j<this.noHosts; j++){
				System.out.print(hosts[j]+"\t");
			}
			System.out.println();
			for (int i=0; i<this.noParasites; i++){
				double aValues[] = solver.getValues(a[i]);
				System.out.print(parasites[i]+"\t");
				for (int j=0; j<this.noHosts; j++){
					if (aValues[j]>=0.5){
						associations++;
					}
					if (this.parasiteTree.getNode(parasites[i]).isLeaf() && this.hostTree.getNode(hosts[j]).isLeaf() && aValues[j]>=0.5){
						System.out.print("*");
					}
					System.out.print((aValues[j]<0.5?"0":"1")+"\t");
				}
				System.out.println();
			}
			System.out.println(associations + " associations");
			
			System.out.print("#COSPECIATION MATRIX:\n");
			System.out.print("\t");
			for (int j=0; j<this.noHosts; j++){
				System.out.print(hosts[j]+"\t");
			}
			System.out.println();
			for (int i=0; i<this.noParasites; i++){
				double cValues[] = solver.getValues(c[i]);
				System.out.print(parasites[i]+"\t");
				for (int j=0; j<this.noHosts; j++){
					System.out.print(Math.round(cValues[j])+"\t");
				}
				System.out.println();
			}
			System.out.println(cospeciations + " cospeciations");
			
			if (this.parameters.isHostswitch()){
				System.out.print("#HOSTSWITCH MATRIX:\n");
				System.out.print("\t");
				for (int j=0; j<this.noHosts; j++){
					System.out.print(hosts[j]+"\t");
				}
				System.out.println();
				for (int i=0; i<this.noParasites; i++){
					double hValues[] = solver.getValues(h[i]);
					System.out.print(parasites[i]+"\t");
					for (int j=0; j<this.noHosts; j++){
						System.out.print((hValues[j]<0.5?"0":"1")+"\t");
					}
					System.out.println();
				}
				System.out.println(hostswitches + " hostswitches");
			}
			
			System.out.print("#DESCENDANT ASSOCIATION MATRIX:\n");
			System.out.print("\t");
			for (int j=0; j<this.noHosts; j++){
				System.out.print(hosts[j]+"\t");
			}
			System.out.println();
			for (int i=0; i<this.noParasites; i++){
				double sValues[] = solver.getValues(s[i]);
				System.out.print(parasites[i]+"\t");
				for (int j=0; j<this.noHosts; j++){
					System.out.print(Math.round(sValues[j])+"\t");
				}
				System.out.println();
			}
			
			System.out.print("#EXCLUSIVE ASSOCIATION MATRIX:\n");
			System.out.print("\t");
			for (int j=0; j<this.noHosts; j++){
				System.out.print(hosts[j]+"\t");
			}
			System.out.println();
			for (int i=0; i<this.noParasites; i++){
				System.out.print(parasites[i]+"\t");
				for (int j=0; j<this.noHosts; j++){
					try {
						double eValue = solver.getValue(e[i][j]);
						System.out.print(Math.round(eValue)+"\t");
					} catch (Exception e) {
						System.out.print("-\t");
					}
				}
				System.out.println();
			}
			
			System.out.print("#SCORE: "+solver.getObjValue()+"\n");
			System.out.print("\n");
		}
	}

	private void justifyBranchlengths(Node node, Tree tree, IloNumVar[] f, IloNumVar[] t) throws UnknownObjectException, IloException {
		int indexParent = node.getNumber();
		for (Node child : node.getChildren()){
			justifyBranchlengths(child, tree, f, t);
			int indexChild = child.getNumber();
			Edge edge = tree.getEdge(node.getId(), child.getId());
//			double factor = solver.getValue(f[indexChild]);
//			edge.setJustifiedLength(edge.getLength()*factor);
			double timeParent = Math.round(solver.getValue(t[indexParent]));
			double timeChild = Math.round(solver.getValue(t[indexChild]));
			double length = (timeParent - timeChild)/this.blFactor;
			edge.setScaledLength(length);
		}
	}
	
	public static void writeFile(String content, String file, boolean append) throws IOException{
		BufferedWriter bw = new BufferedWriter(new FileWriter(file, append));
		bw.write(content);
		bw.close();
	}
	
}
